/*
 * robot.c
 *
 *  Created on: 19.07.2016
 *      Author: AndreR
 */

#include "robot.h"
#include "commands.h"
#include "struct_ids.h"
#include "skills.h"
#include "ctrl.h"
#include "util/sys_time.h"
#include "network.h"
#include "main/version.h"
#include "util/boot.h"
#include "main/ctrl_tigga.h"

Robot robot = {
	.botLogging = { 1, 1 },
	.misc = { 0xFF, 0xFF },
};

#define ROBOT_MODE_IDLE			0
#define ROBOT_MODE_READY		1
#define ROBOT_MODE_LOW_POWER	2
#define ROBOT_MODE_TEST			3

#define ROBOT_EVENT_BAT_EMPTY		0
#define ROBOT_EVENT_VISION_RECEIVED	1
#define ROBOT_EVENT_ENTER_TEST		2
#define ROBOT_EVENT_EXIT_TEST		3
#define ROBOT_EVENT_RESET			4

static void clearUpdateFlags();
static void doLogging();
static void sendMatchFeedback();
static void parseMatchCtrl();
static void connectionCheck();
static void processEvent(uint32_t event);
static void stateWatchdog();
static void handleDataAcquisitionMode();

static const ConfigFileDesc configFileDescLogging =
	{ SID_CFG_CTRL_LOGGING, 3, "ctrl/logs", 2, (ElementDesc[]) {
		{ UINT8, "in_state_out", "", "Input/State/Output" },
		{ UINT8, "specific", "", "Specific Ctrl. Data" },
	} };

static void initLogFields()
{
	memset(&robot.sensors, 0, sizeof(robot.sensors));
	memset(&robot.state, 0, sizeof(robot.state));
	memset(&robot.skillOutput, 0, sizeof(robot.skillOutput));
	memset(&robot.ctrlOutput, 0, sizeof(robot.ctrlOutput));
	memset(&robot.ctrlRef, 0, sizeof(robot.ctrlRef));
	memset(&robot.netStats, 0, sizeof(robot.netStats));

	robot.sensors.header.type = SID_SENSORS;
	robot.sensors.header.length = sizeof(RobotSensors);
	robot.state.header.type = SID_CTRL_STATE;
	robot.state.header.length = sizeof(RobotCtrlState);
	robot.skillOutput.header.type = SID_SKILL_OUTPUT;
	robot.skillOutput.header.length = sizeof(SkillOutput);
	robot.ctrlOutput.header.type = SID_CTRL_OUTPUT;
	robot.ctrlOutput.header.length = sizeof(RobotCtrlOutput);
	robot.ctrlRef.header.type = SID_CTRL_REF;
	robot.ctrlRef.header.length = sizeof(RobotCtrlReference);
	robot.netStats.header.type = SID_NETSTATS;
	robot.netStats.header.length = sizeof(NetStats);
}

void RobotInit()
{
	robot.mode = ROBOT_MODE_IDLE;
	robot.enabledSystems = ROBOT_SYSTEM_BUZZER | ROBOT_SYSTEM_LEDS;

	chMBObjectInit(&robot.eventQueue, robot.eventQueueData, ROBOT_EVENT_QUEUE_SIZE);

	LagElementPT1Init(&robot.accZLag, 1, 0.5f, 0.001f);
	robot.accZLag.state[0] = 20.0f;
	robot.accZLag.state[1] = 20.0f;
	robot.accZLag.state[2] = 20.0f;
	robot.accZLag.state[3] = 20.0f;

	robot.ext.lastMatchCtrlTime = SysTimeUSec()-5000000;

	initLogFields();

	FlashFSOpenOrCreate("botStatus", 0, &robot.misc, sizeof(ConfigMisc), &robot.pMiscConfig);

	robot.pBotLoggingConfig = ConfigOpenOrCreate(&configFileDescLogging, &robot.botLogging, sizeof(BotLogging), 0, 1);

	SkillsInit();
	CtrlInit();
}

void RobotTask(void* params)
{
	(void)params;

	chRegSetThreadName("Robot");

	uint32_t event;
	systime_t prev = chVTGetSystemTimeX();
	systime_t next = prev+US2ST(1000);

	while(1)
	{
		robot.sensors.header.timestamp = SysTimeUSec();

		// fetch updated sensor data from all subsystems
		RobotImplUpdateSensors(&robot.sensors);

		// check connection to BS and command sender
		connectionCheck();

		// get skill data, parse flags, read vision pos (if available)
		parseMatchCtrl();

		// run state estimation with sensor input
		if(robot.enabledSystems & ROBOT_SYSTEM_STATE_EST)
			CtrlStateEstimation(&robot.sensors, &robot.ctrlOutput, &robot.state);

		robot.state.header.timestamp = SysTimeUSec();

		// execute skill system with updated state
		if(robot.enabledSystems & ROBOT_SYSTEM_SKILLS)
			SkillsExecute(&robot.sensors, &robot.state, &robot.ctrlOutput, &robot.ctrlRef, &robot.skillOutput);

		robot.skillOutput.header.timestamp = SysTimeUSec();

		// run control algorithms for driving
		if(robot.enabledSystems & ROBOT_SYSTEM_CONTROL)
			CtrlControl(&robot.state, &robot.skillOutput.drive, &robot.ctrlOutput, &robot.ctrlRef);

		robot.ctrlOutput.header.timestamp = SysTimeUSec();
		robot.ctrlRef.header.timestamp = SysTimeUSec();

		// Apply all actuator outputs
		if(robot.enabledSystems & ROBOT_SYSTEM_DRIVE)
			RobotImplMotorOutput(&robot.ctrlOutput);

		if(robot.enabledSystems & ROBOT_SYSTEM_KICKER)
			RobotImplKickerOutput(&robot.skillOutput.kicker);

		if(robot.enabledSystems & ROBOT_SYSTEM_DRIBBLER)
			RobotImplDribblerOutput(&robot.skillOutput.dribbler);

		// do sensor/state/output logging
		doLogging();

		// send high sample rate acquisition mode data
		handleDataAcquisitionMode();

		// send match feedback to Sumatra
		sendMatchFeedback();

		// battery check
		if(robot.sensors.power.batEmpty)
			chMBPost(&robot.eventQueue, ROBOT_EVENT_BAT_EMPTY, TIME_IMMEDIATE);

		// state watchdog (checks Z Acc)
		stateWatchdog();

		// clear all updated flags in sensor data
		clearUpdateFlags();

		// process all events
		while(chMBFetch(&robot.eventQueue, (msg_t*)&event, TIME_IMMEDIATE) == MSG_OK)
		{
			processEvent(event);
		}

		// wait until next loop
		prev = chThdSleepUntilWindowed(prev, next);
		next += US2ST(1000);
	}
}

void RobotSetHardwareId(uint8_t hwId)
{
	robot.misc.hardwareId = hwId;
	FlashFSWrite(&robot.pMiscConfig, &robot.misc, sizeof(ConfigMisc));
}

void RobotTestModeEnable(uint8_t enable)
{
	if(enable)
		chMBPost(&robot.eventQueue, ROBOT_EVENT_ENTER_TEST, TIME_INFINITE);
	else
		chMBPost(&robot.eventQueue, ROBOT_EVENT_EXIT_TEST, TIME_INFINITE);

	chThdSleepMilliseconds(10);
}

void RobotTestModeSetSystems(uint32_t enabledSystems)
{
	if(robot.mode == ROBOT_MODE_TEST)
		robot.enabledSystems = enabledSystems;

	chThdSleepMilliseconds(10);
}

static void stop()
{
	initLogFields();

	CtrlCheckControllerChange(&robot.ctrlOutput);

	RobotImplMotorOutput(&robot.ctrlOutput);
	RobotImplKickerOutput(&robot.skillOutput.kicker);
	RobotImplDribblerOutput(&robot.skillOutput.dribbler);
}

void RobotTestModeStopAll()
{
	if(robot.mode != ROBOT_MODE_TEST)
		return;

	RobotTestModeSetSystems(0);

	stop();

	RobotImplBuzzerPlay(0);
	RobotImplKickerAutoCharge(0);
	RobotImplSetLEDs(0);
}

static void processEvent(uint32_t event)
{
	if(robot.mode == ROBOT_MODE_LOW_POWER)
		return;

	if(event == ROBOT_EVENT_BAT_EMPTY)
	{
		if(robot.mode == ROBOT_MODE_READY)
			robot.enabledSystems = ROBOT_SYSTEM_STATE_EST;
		else
			robot.enabledSystems = 0;

		stop();

		RobotImplKickerAutoCharge(0);
		RobotImplBuzzerPlay(BUZZ_BEEP_FAST);

		robot.mode = ROBOT_MODE_LOW_POWER;
	}

	if(event == ROBOT_EVENT_ENTER_TEST)
	{
		robot.mode = ROBOT_MODE_TEST;
	}

	if(robot.mode == ROBOT_MODE_TEST && event == ROBOT_EVENT_EXIT_TEST)
	{
		robot.mode = ROBOT_MODE_IDLE;

		robot.enabledSystems = ROBOT_SYSTEM_BUZZER | ROBOT_SYSTEM_LEDS;

		ctrl.forceControllerReload = 1;

		stop();
	}

	if(robot.mode == ROBOT_MODE_IDLE && event == ROBOT_EVENT_VISION_RECEIVED)
	{
		robot.mode = ROBOT_MODE_READY;

		robot.enabledSystems = ROBOT_SYSTEM_ALL_MASK;
	}

	if(event == ROBOT_EVENT_RESET)
	{
		robot.mode = ROBOT_MODE_IDLE;

		robot.enabledSystems = ROBOT_SYSTEM_BUZZER | ROBOT_SYSTEM_LEDS;

		ctrl.forceControllerReload = 1;

		stop();
	}
}

static void doLogging()
{
	if(robot.botLogging.ctrlCommon)
	{
		RobotImplWriteLogMsg(&robot.sensors);
		RobotImplWriteLogMsg(&robot.state);
		RobotImplWriteLogMsg(&robot.skillOutput);
		RobotImplWriteLogMsg(&robot.ctrlOutput);
		RobotImplWriteLogMsg(&robot.ctrlRef);
	}

	if(robot.botLogging.ctrlSpecific)
	{
		if(ctrl.pInstance && ctrl.pInstance->pLog)
			RobotImplWriteLogMsg(ctrl.pInstance->pLog);
	}

	static uint8_t netStatCnt = 0;
	if(netStatCnt == 100)
	{
		netStatCnt = 0;
		RobotImplWriteLogMsg(&robot.netStats);
	}
	++netStatCnt;
}

static uint16_t gatherFeatures()
{
	uint16_t features = 0;

	if(botParams.v2016)
		features |= SYSTEM_MATCH_FEEDBACK_FEATURE_V2016;

	if(botParams.extInstalled)
		features |= SYSTEM_MATCH_FEEDBACK_FEATURE_EXT_BOARD;

	if((robot.sensors.kicker.flags & KICKER_FLAGS_DMG_BARRIER) == 0)
		features |= SYSTEM_MATCH_FEEDBACK_FEATURE_BARRIER;

	if(robot.sensors.kicker.flags & KICKER_FLAGS_V2017)
		features |= SYSTEM_MATCH_FEEDBACK_FEATURE_KICKER_V2017;

	if(robot.sensors.power.exhausted == 0)
		features |= SYSTEM_MATCH_FEEDBACK_FEATURE_ENERGETIC;

	if(robot.sensors.power.batEmpty == 1)
		return features;

	const uint32_t moveSystems = ROBOT_SYSTEM_SKILLS | ROBOT_SYSTEM_DRIVE | ROBOT_SYSTEM_CONTROL;
	if((robot.enabledSystems & moveSystems) == moveSystems)
		features |= SYSTEM_MATCH_FEEDBACK_FEATURE_MOVE;

	const uint32_t dribbleSystems = ROBOT_SYSTEM_SKILLS | ROBOT_SYSTEM_DRIBBLER;
	if((robot.enabledSystems & dribbleSystems) == dribbleSystems && robot.sensors.dribbler.overheated == 0)
		features |= SYSTEM_MATCH_FEEDBACK_FEATURE_DRIBBLE;

	const uint32_t kickSystems = ROBOT_SYSTEM_SKILLS | ROBOT_SYSTEM_KICKER;
	if((robot.enabledSystems & kickSystems) == kickSystems)
	{
		if((robot.sensors.kicker.flags & KICKER_FLAGS_DMG_STRAIGHT) == 0)
			features |= SYSTEM_MATCH_FEEDBACK_FEATURE_STRAIGHT;

		if((robot.sensors.kicker.flags & KICKER_FLAGS_DMG_CHIP) == 0)
			features |= SYSTEM_MATCH_FEEDBACK_FEATURE_CHIP;

		if((robot.sensors.kicker.flags & KICKER_FLAGS_DMG_CHARGE) == 0)
			features |= SYSTEM_MATCH_FEEDBACK_FEATURE_CHARGE;
	}

	return features;
}

static void sendMatchFeedback()
{
	if(robot.state.posUpdated)
		robot.feedbackStateUpdated[0] = 1;
	if(robot.state.velUpdated)
		robot.feedbackStateUpdated[1] = 1;

	// check if there are other packets to be send, if not => send a feedback
	uint8_t txActive = network.queue.txFifo.numPackets > 0 || network.queue.txBufUsed > 0;

	// also use a timeout of 100ms to send a feedback at at least 10Hz
	if(txActive && (SysTimeUSecRaw() - robot.lastLogTime) < 100000)
		return;

	robot.lastLogTime = SysTimeUSecRaw();

	SystemMatchFeedback feedback;
	memset(&feedback, 0, sizeof(SystemMatchFeedback));

	feedback.batteryLevel = (uint16_t)(robot.sensors.power.voltage*1000.0f);
	feedback.kickerLevel = robot.sensors.kicker.level;
	feedback.barrierKickCounter = (robot.sensors.ir.interrupted << 7) | (robot.sensors.kicker.kickCounter & 0x7F);
	feedback.dribblerSpeed = robot.sensors.dribbler.speed;
	feedback.dribblerTemp = (uint8_t)(robot.sensors.dribbler.temp*2.0f);
	feedback.hardwareId = robot.misc.hardwareId;
	feedback.features = gatherFeatures() | (robot.mode << 12);

	const float dt = 0.02f;

	if(robot.feedbackStateUpdated[0])
	{
		robot.feedbackStateUpdated[0] = 0;

		float futurePos[2];
		futurePos[0] = robot.state.pos[0] + dt*robot.state.vel[0] + 0.5f*dt*dt*robot.state.acc[0];
		futurePos[1] = robot.state.pos[1] + dt*robot.state.vel[1] + 0.5f*dt*dt*robot.state.acc[1];

		feedback.curPosition[0] = (int16_t)(futurePos[0]*1000.0f);
		feedback.curPosition[1] = (int16_t)(futurePos[1]*1000.0f);
		feedback.curPosition[2] = (int16_t)(robot.state.pos[2]*1000.0f);
	}
	else
	{
		feedback.curPosition[0] = SYSTEM_MATCH_FEEDBACK_UNUSED_FIELD;
		feedback.curPosition[1] = SYSTEM_MATCH_FEEDBACK_UNUSED_FIELD;
		feedback.curPosition[2] = SYSTEM_MATCH_FEEDBACK_UNUSED_FIELD;
	}

	if(robot.feedbackStateUpdated[1])
	{
		robot.feedbackStateUpdated[1] = 0;

		float futureVel[2];
		futureVel[0] = robot.state.vel[0] + dt*robot.state.acc[0];
		futureVel[1] = robot.state.vel[1] + dt*robot.state.acc[1];

		feedback.curVelocity[0] = (int16_t)(futureVel[0]*1000.0f);
		feedback.curVelocity[1] = (int16_t)(futureVel[1]*1000.0f);
		feedback.curVelocity[2] = (int16_t)(robot.state.vel[2]*1000.0f);
	}
	else
	{
		feedback.curVelocity[0] = SYSTEM_MATCH_FEEDBACK_UNUSED_FIELD;
		feedback.curVelocity[1] = SYSTEM_MATCH_FEEDBACK_UNUSED_FIELD;
		feedback.curVelocity[2] = SYSTEM_MATCH_FEEDBACK_UNUSED_FIELD;
	}

	PacketHeader header;
	header.section = SECTION_SYSTEM;
	header.cmd = CMD_SYSTEM_MATCH_FEEDBACK;
	NetworkSendPacket(&header, (uint8_t*) &feedback, sizeof(SystemMatchFeedback));

	RobotImplExtSendMatchFeedback(&feedback);
}

static void clearUpdateFlags()
{
	robot.sensors.gyr.updated = 0;
	robot.sensors.acc.updated = 0;
	robot.sensors.vision.updated = 0;
	robot.state.posUpdated = 0;
	robot.state.velUpdated = 0;
}

static void parseMatchCtrl()
{
	static uint8_t lastSong;

	// prefer match ctrl cmd from primary (nRF24) interface
	uint32_t matchCtrlTime = network.matchCtrlTime;
	SystemMatchCtrl* pCtrl = &network.lastMatchCtrlCmd;

	// only take external command interface if sumatra is offline
	if(!robot.sumatraOnline && robot.ext.active)
	{
		pCtrl = &robot.ext.matchCtrl;
		matchCtrlTime = robot.ext.lastMatchCtrlTime;
	}

	if(robot.sensors.vision.time != matchCtrlTime)
	{
		if(pCtrl->curPosition[0] != SYSTEM_MATCH_CTRL_UNUSED_FIELD)
		{
			uint32_t delay = pCtrl->posDelay;
			delay = (delay*25000)/100;	// delay now in [us]

			robot.sensors.vision.pos[0] = pCtrl->curPosition[0]*0.001f;
			robot.sensors.vision.pos[1] = pCtrl->curPosition[1]*0.001f;
			robot.sensors.vision.pos[2] = pCtrl->curPosition[2]*0.001f;
			robot.sensors.vision.time = matchCtrlTime;
			robot.sensors.vision.delay = delay;
			robot.sensors.vision.camId = pCtrl->camId & 0x7F;
			robot.sensors.vision.updated = 1;
			robot.sensors.vision.noVision = 0;

			chMBPost(&robot.eventQueue, ROBOT_EVENT_VISION_RECEIVED, TIME_IMMEDIATE);
		}

		if(pCtrl->camId & SYSTEM_MATCH_CTRL_CAM_ID_FLAG_NO_VISION)
		{
			robot.sensors.vision.time = matchCtrlTime;
			robot.sensors.vision.updated = 0;
			robot.sensors.vision.noVision = 1;

			chMBPost(&robot.eventQueue, ROBOT_EVENT_VISION_RECEIVED, TIME_IMMEDIATE);
		}
	}

	if(robot.enabledSystems & ROBOT_SYSTEM_KICKER)
	{
		if(pCtrl->flags & SYSTEM_MATCH_CTRL_FLAGS_KICKER_AUTOCHG)
			RobotImplKickerAutoCharge(1);
		else
			RobotImplKickerAutoCharge(0);
	}

	uint8_t song = pCtrl->flags & SYSTEM_MATCH_CTRL_FLAGS_SONG_MASK;

	if(song != lastSong && (robot.enabledSystems & ROBOT_SYSTEM_BUZZER))
		RobotImplBuzzerPlay(song);

	lastSong = song;

	uint8_t leds = (pCtrl->flags & SYSTEM_MATCH_CTRL_FLAGS_LED_MASK) >> 4;
	if(robot.enabledSystems & ROBOT_SYSTEM_LEDS)
		RobotImplSetLEDs(leds);

	if(pCtrl->flags & SYSTEM_MATCH_CTRL_FLAGS_STRICT_VEL_LIMIT)
		robot.skillOutput.drive.limits.strictVelLimit = 1;
	else
		robot.skillOutput.drive.limits.strictVelLimit = 0;

	if(robot.enabledSystems & ROBOT_SYSTEM_SKILLS)
	{
		if(robot.sumatraOnline == 0 && robot.ext.active == 0)
			SkillsSetInput(0, 0);
		else
			SkillsSetInput(pCtrl->skillId, pCtrl->skillData);
	}
}

static void sendVersionInfo()
{
	PacketHeader header;
	header.section = SECTION_SYSTEM;
	header.cmd = CMD_SYSTEM_VERSION;

	SystemVersion version;
	version.version = VersionGet();
	version.crc32 = BootGetApplicationCRC32();

	NetworkSendPacketReliable(&header, &version, sizeof(SystemVersion));
}

static void connectionCheck()
{
	if(network.avgRxPeriod.value > 0.08f && robot.bsOnline)
	{
		robot.bsOnline = 0;
	}

	if(network.avgRxPeriod.value < 0.08f && !robot.bsOnline)
	{
		robot.bsOnline = 1;
	}

	if(SysTimeUSec() - network.matchCtrlTime > 1000000)
	{
		if(robot.sumatraOnline)
		{
			robot.sumatraOnline = 0;

			if(robot.enabledSystems & ROBOT_SYSTEM_BUZZER)
				RobotImplBuzzerPlay(BUZZ_DOWN50);
		}
	}
	else
	{
		if(robot.sumatraOnline == 0)
		{
			robot.sumatraOnline = 1;

			sendVersionInfo();

			if(robot.enabledSystems & ROBOT_SYSTEM_BUZZER)
				RobotImplBuzzerPlay(BUZZ_UP50);
		}
	}

	if(SysTimeUSec() - robot.ext.lastMatchCtrlTime > 1000000)
	{
		if(robot.ext.active)
		{
			robot.ext.active = 0;

			if(robot.enabledSystems & ROBOT_SYSTEM_BUZZER)
				RobotImplBuzzerPlay(BUZZ_DOWN50);
		}
	}
	else
	{
		if(robot.ext.active == 0)
		{
			robot.ext.active = 1;

			if(robot.enabledSystems & ROBOT_SYSTEM_BUZZER)
				RobotImplBuzzerPlay(BUZZ_UP50);
		}
	}

	robot.wifiStat.bsOnline = robot.bsOnline;
	robot.wifiStat.sumatraOnline = robot.sumatraOnline;
	robot.wifiStat.updateFreq = 1.0f/network.avgRxPeriod.value;
	robot.wifiStat.visionDelay = robot.sensors.vision.delay;
	robot.wifiStat.rssi = network.avgRssi.value;

	robot.netStats.header.timestamp = SysTimeUSec();
	robot.netStats.rxPeriod = network.avgRxPeriod.value;
	robot.netStats.rssi = network.avgRssi.value;
	robot.netStats.rxPackets = network.queue.stats.rfIO.rxPackets;
	robot.netStats.rxBytes = network.queue.stats.rfIO.rxBytes;
	robot.netStats.txPackets = network.queue.stats.rfIO.txPackets;
	robot.netStats.txBytes = network.queue.stats.rfIO.txBytes;
	robot.netStats.rxPacketsLost = network.queue.stats.rfFeedback.rxPacketsLost;
	robot.netStats.rxHLPacketsLost = network.queue.stats.queueFeedback.rxPacketsLost;
	robot.netStats.txHLPacketsLost = network.queue.stats.queueFeedback.txPacketsLost;
}


static void stateWatchdog()
{
	if(robot.sensors.acc.updated)
		robot.accZ = LagElementPT1Process(&robot.accZLag, robot.sensors.acc.linAcc[2]);

	if(robot.accZ < 3.0f)
	{
		chMBPost(&robot.eventQueue, ROBOT_EVENT_RESET, TIME_IMMEDIATE);
	}

	int posNaN = isnanf(robot.state.pos[0]) || isnanf(robot.state.pos[1]) || isnanf(robot.state.pos[2]);
	int velNaN = isnanf(robot.state.vel[0]) || isnanf(robot.state.vel[1]) || isnanf(robot.state.vel[2]);
	int accNaN = isnanf(robot.state.acc[0]) || isnanf(robot.state.acc[1]) || isnanf(robot.state.acc[2]);

	if(posNaN || velNaN || accNaN)
	{
		chMBPost(&robot.eventQueue, ROBOT_EVENT_RESET, TIME_IMMEDIATE);
	}
}

static void handleDataAcquisitionMode()
{
	// check if there are other packets to be send, if not => send a feedback
	uint8_t txActive = network.queue.txFifo.numPackets > 0 || network.queue.txBufUsed > 0;
	if(txActive)
		return;

	switch(robot.dataAcquisitionMode)
	{
		case ACQ_MODE_MOTOR_MODEL:
		{
			DataAcqMotorModel mm;
			mm.timestamp = robot.sensors.motorVol.time;
			for(uint8_t i = 0; i < 4; i++)
			{
				mm.motVol[i] = (int16_t)(robot.sensors.motorVol.vol[i]*1000.0f);
				mm.motVel[i] = (int16_t)(robot.sensors.enc.vel[i]*40.0f);
			}

			PacketHeader header;
			header.section = SECTION_DATA_ACQ;
			header.cmd = CMD_DATA_ACQ_MOTOR_MODEL;

			NetworkSendPacket(&header, &mm, sizeof(DataAcqMotorModel));
		}
		break;
		case ACQ_MODE_BOT_MODEL:
		{
			DataAcqBotModel bm;
			bm.timestamp = robot.ctrlOutput.header.timestamp;

			float outVel[3];
			CtrlUtilMotorVelToLocalVel(robot.ctrlOutput.motorVel, outVel);

			bm.outVel[0] = (int16_t)(outVel[0]*1000.0f);
			bm.outVel[1] = (int16_t)(outVel[1]*1000.0f);
			bm.outVel[2] = (int16_t)(outVel[2]*100.0f);

			for(uint8_t i = 0; i < 3; i++)
				bm.visPos[i] = (int16_t)(robot.sensors.vision.pos[i]*1000.0f);

			bm.visionTime = robot.sensors.vision.time - robot.sensors.vision.delay;

			PacketHeader header;
			header.section = SECTION_DATA_ACQ;
			header.cmd = CMD_DATA_ACQ_BOT_MODEL;

			NetworkSendPacket(&header, &bm, sizeof(DataAcqBotModel));
		}
		break;
		case ACQ_MODE_DELAYS:
		{
			DataAcqDelays de;
			de.timestamp = robot.sensors.header.timestamp;
			de.visPosW = (int16_t)(robot.sensors.vision.pos[2]*1000.0f);
			de.visionTime = robot.sensors.vision.time - robot.sensors.vision.delay;

			float outVel[3];
			CtrlUtilMotorVelToLocalVel(robot.ctrlOutput.motorVel, outVel);

			de.outVelW = (int16_t)(outVel[2]*1000.0f);
			de.gyrVel = (int16_t)(robot.sensors.gyr.rotVel[2]*1000.0f);

			PacketHeader header;
			header.section = SECTION_DATA_ACQ;
			header.cmd = CMD_DATA_ACQ_DELAYS;

			NetworkSendPacket(&header, &de, sizeof(DataAcqDelays));
		}
		break;
		case ACQ_MODE_BOT_MODEL_V2:
		{
			DataAcqBotModelV2 bm2;

			bm2.timestamp = robot.ctrlOutput.header.timestamp;

			float encVel[3];
			CtrlUtilMotorVelToLocalVel(robot.sensors.enc.vel, encVel);

			bm2.encVel[0] = (int16_t)(encVel[0]*1000.0f);
			bm2.encVel[1] = (int16_t)(encVel[1]*1000.0f);
			bm2.encVel[2] = (int16_t)(encVel[2]*100.0f);

			float localStateVel[2];
			CtrlUtilTurnGlobal2Local(robot.state.pos[2], robot.state.vel[0], robot.state.vel[1], localStateVel, localStateVel+1);

			bm2.stateVel[0] = (int16_t)(localStateVel[0]*1000.0f);
			bm2.stateVel[1] = (int16_t)(localStateVel[1]*1000.0f);
			bm2.stateVel[2] = (int16_t)(robot.state.vel[2]*100.0f);

			float localForce[3];
			CtrlUtilMotorTorqueToLocalForce(robot.ctrlOutput.motorTorque, localForce);

			bm2.outForce[0] = (int16_t)(localForce[0]*100.0f);
			bm2.outForce[1] = (int16_t)(localForce[1]*100.0f);
			bm2.outForce[2] = (int16_t)(localForce[2]*1000.0f);

			uint16_t effXY = (uint16_t)(ctrlTigga.configModel.efficiencyXY*4096.0f);
			uint16_t effW = (uint16_t)(ctrlTigga.configModel.efficiencyW*4096.0f);

			bm2.efficiency[0] = effXY & 0xFF;
			bm2.efficiency[1] = ((effXY & 0xF00) >> 8) | ((effW & 0x0F) << 4);
			bm2.efficiency[2] = (effW & 0xFF0) >> 4;

			uint8_t modeXY = 0;

			if(robot.skillOutput.drive.modeXY == DRIVE_MODE_LOCAL_FORCE)
			{
				modeXY = 1;
			}
			else if(robot.skillOutput.drive.modeXY == DRIVE_MODE_LOCAL_VEL ||
					robot.skillOutput.drive.modeXY == DRIVE_MODE_GLOBAL_POS ||
					robot.skillOutput.drive.modeXY == DRIVE_MODE_GLOBAL_POS_ASYNC)
			{
				const float* pTrajVel = robot.ctrlRef.trajVelLocal;
				const float* pTrajAcc = robot.ctrlRef.trajAccLocal;
				const float trajVelAbs = sqrtf(pTrajVel[0]*pTrajVel[0] + pTrajVel[1]*pTrajVel[1]);
				const float trajAccAbs = sqrtf(pTrajAcc[0]*pTrajAcc[0] + pTrajAcc[1]*pTrajAcc[1]);

				if(trajAccAbs < 1e-3f)
				{
					if(trajVelAbs < 1e-3f)
					{
						// stopped
						modeXY = 2;
					}
					else
					{
						// plateau
						modeXY = 3;
					}
				}
				else
				{
					float accDotVel;
					arm_dot_prod_f32(pTrajAcc, pTrajVel, 2, &accDotVel);
					if(accDotVel > 0)
					{
						// accelerating
						modeXY = 4;
					}
					else
					{
						// decelerating
						modeXY = 5;
					}
				}
			}

			uint8_t modeW = 0;

			if(robot.skillOutput.drive.modeW == DRIVE_MODE_LOCAL_FORCE)
			{
				modeW = 1;
			}
			else if(robot.skillOutput.drive.modeW == DRIVE_MODE_LOCAL_VEL ||
					robot.skillOutput.drive.modeW == DRIVE_MODE_GLOBAL_POS ||
					robot.skillOutput.drive.modeW == DRIVE_MODE_GLOBAL_POS_ASYNC)
			{
				const float trajVel = robot.ctrlRef.trajVelLocal[2];
				const float trajAcc = robot.ctrlRef.trajAccLocal[2];
				const float trajVelAbs = fabsf(trajVel);
				const float trajAccAbs = fabsf(trajAcc);

				if(trajAccAbs < 1e-3f)
				{
					if(trajVelAbs < 1e-3f)
					{
						// stopped
						modeW = 2;
					}
					else
					{
						// plateau
						modeW = 3;
					}
				}
				else
				{
					if(trajVel*trajAcc > 0)
					{
						// accelerating
						modeW = 4;
					}
					else
					{
						// decelerating
						modeW = 5;
					}
				}
			}

			bm2.mode = modeXY | (modeW << 4);

			PacketHeader header;
			header.section = SECTION_DATA_ACQ;
			header.cmd = CMD_DATA_ACQ_BOT_MODEL_V2;

			NetworkSendPacket(&header, &bm2, sizeof(DataAcqBotModelV2));
		}
		break;
	}
}
